XIncludeFile "GPUCalc.pbi"

Procedure WaitForLastCommand(*context.GPUCalcContext)
  Protected result

  result = 1

  While Not *context\hasFinishedLastCommand()
    gpuCalcYield()
  Wend

  If Not *context\wasLastCommandSuccessful()
    Select *context\getLastErrorCode()
      Case #GPUCALC_ERR_NO_ERROR
        PrintN("#GPUCALC_ERR_NO_ERROR")
      Case #GPUCALC_ERR_RENDERINGCONTEXT
        PrintN("#GPUCALC_ERR_RENDERINGCONTEXT")
      Case #GPUCALC_ERR_MULTITEXTURING
        PrintN("#GPUCALC_ERR_MULTITEXTURING")
      Case #GPUCALC_ERR_SHADERS
        PrintN("#GPUCALC_ERR_SHADERS")
      Case #GPUCALC_ERR_FRAMEBUFFEROBJECTS
        PrintN("#GPUCALC_ERR_FRAMEBUFFEROBJECTS")
      Case #GPUCALC_ERR_VERTEXBUFFEROBJECTS
        PrintN("#GPUCALC_ERR_VERTEXBUFFEROBJECTS")
      Case #GPUCALC_ERR_COMPILE_VERTEX
        PrintN("#GPUCALC_ERR_COMPILE_VERTEX")
        PrintN(PeekS(*context\getCompilerErrorVertex()))
      Case #GPUCALC_ERR_COMPILE_FRAGMENT
        PrintN("#GPUCALC_ERR_COMPILE_FRAGMENT")
        PrintN(PeekS(*context\getCompilerErrorFragment()))
      Case #GPUCALC_ERR_LINK_SHADERS
        PrintN("#GPUCALC_ERR_LINK_SHADERS")
        PrintN(PeekS(*context\getLinkerError()))
      Case #GPUCALC_ERR_FLOATTEXTURE
        PrintN("#GPUCALC_ERR_FLOATTEXTURE")
      Case #GPUCALC_ERR_INTEGERTEXTURE
        PrintN("#GPUCALC_ERR_INTEGERTEXTURE")
      Case #GPUCALC_ERR_INVALIDDATATYPE
        PrintN("#GPUCALC_ERR_INVALIDDATATYPE")
    EndSelect
    result = 0
  EndIf

  ProcedureReturn result
EndProcedure

Procedure.q AbsI(val.q)
  If val >= 0
    ProcedureReturn val
  EndIf
  ProcedureReturn val * -1
EndProcedure

Procedure.q Clamp(val.q, min.q, max.q)
  If val > max
    val = max
  ElseIf val < min
    val = min
  EndIf
  
  ProcedureReturn val
EndProcedure

Procedure.d CompareDatasets(datatype.i, elements.i, *dataset1.DATASETMEMORY, *dataset2.DATASETMEMORY)
  Protected result.d = 0.0
  Protected k
  
  Select Datatype
    Case #GPUCALC_TYPE_FLOAT32
      For k = 0 To elements - 1
        difference = Abs(*dataset1\float[k] - *dataset2\float[k])
        If difference > result
          result = difference
        EndIf
      Next k
      
    Case #GPUCALC_TYPE_UINT32
      For k = 0 To elements - 1
        difference = AbsI(*dataset1\uint[k] - *dataset2\uint[k])
        If difference > result
          result = difference
        EndIf
      Next k
      
    Case #GPUCALC_TYPE_UINT16
      For k = 0 To elements - 1
        difference = AbsI(*dataset1\ushort[k] - *dataset2\ushort[k])
        If difference > result
          result = difference
        EndIf
      Next k
      
    Case #GPUCALC_TYPE_UINT8
      For k = 0 To elements - 1
        difference = AbsI(*dataset1\ubyte[k] - *dataset2\ubyte[k])
        If difference > result
          result = difference
        EndIf
      Next k
      
    Case #GPUCALC_TYPE_INT32
      For k = 0 To elements - 1
        difference = AbsI(*dataset1\int[k] - *dataset2\int[k])
        If difference > result
          result = difference
        EndIf
      Next k
      
    Case #GPUCALC_TYPE_INT16
      For k = 0 To elements - 1
        difference = AbsI(*dataset1\short[k] - *dataset2\short[k])
        If difference > result
          result = difference
        EndIf
      Next k
      
    Case #GPUCALC_TYPE_INT8
      For k = 0 To elements - 1
        difference = AbsI(*dataset1\byte[k] - *dataset2\byte[k])
        If difference > result
          result = difference
        EndIf
      Next k
  EndSelect
  
  ProcedureReturn result
EndProcedure

Procedure FillRandom(datatype.i, *dataset.DATASETMEMORY, elements.i)
  Protected k.i
  
  Select datatype
    Case #GPUCALC_TYPE_FLOAT32
      For k = 0 To elements - 1
        *dataset\float[k] = (Random(1000000) - Random(1000000)) * 0.001
      Next k
    Case #GPUCALC_TYPE_UINT32
      For k = 0 To elements - 1
        *dataset\int[k] = Random($7FFFFFFF)
      Next k
    Case #GPUCALC_TYPE_UINT16
      For k = 0 To elements - 1
        *dataset\short[k] = Random($7FFF)
      Next k
    Case #GPUCALC_TYPE_UINT8
      For k = 0 To elements - 1
        *dataset\byte[k] = Random($7F)
      Next k
    Case #GPUCALC_TYPE_INT32
      For k = 0 To elements - 1
        *dataset\int[k] = Random($7FFFFFFF) - Random($7FFFFFFF)
      Next k
    Case #GPUCALC_TYPE_INT16
      For k = 0 To elements - 1
        *dataset\short[k] = Random($7FFF) - Random($7FFF)
      Next k
    Case #GPUCALC_TYPE_INT8
      For k = 0 To elements - 1
        *dataset\byte[k] = Random($7F) - Random($7F)
      Next k
  EndSelect
EndProcedure

Procedure PrintFirstNResults(datatype.i, elements.i, *a.DATASETMEMORY, *b.DATASETMEMORY, *gpu.DATASETMEMORY, *cpu.DATASETMEMORY)
  Protected k.i
  Protected decimals.i = 12
  
  PrintN(" " + RSet("a", decimals) + " + " + RSet("b", decimals) + " = " + RSet("gpu", decimals) + ", " + RSet("cpu", decimals))
  
  Select datatype
    Case #GPUCALC_TYPE_FLOAT32
      For k = 0 To elements - 1
        PrintN(" " + RSet(StrD(*a\float[k], 3), decimals) + " + " + RSet(StrD(*b\float[k], 3), decimals) + " = " + RSet(StrD(*gpu\float[k], 3), decimals) + ", " + RSet(StrD(*cpu\float[k], 3), decimals))
      Next k
    Case #GPUCALC_TYPE_UINT32
      For k = 0 To elements - 1
        PrintN(" " + RSet(StrU(*a\uint[k], #PB_Long), decimals) + " + " + RSet(StrU(*b\uint[k], #PB_Long), decimals) + " = " + RSet(StrU(*gpu\uint[k], #PB_Long), decimals) + ", " + RSet(StrU(*cpu\uint[k], #PB_Long), decimals))
      Next k
    Case #GPUCALC_TYPE_UINT16
      For k = 0 To elements - 1
        PrintN(" " + RSet(StrU(*a\ushort[k], #PB_Word), decimals) + " + " + RSet(StrU(*b\ushort[k], #PB_Word), decimals) + " = " + RSet(StrU(*gpu\ushort[k], #PB_Word), decimals) + ", " + RSet(StrU(*cpu\ushort[k], #PB_Word), decimals))
      Next k
    Case #GPUCALC_TYPE_UINT8
      For k = 0 To elements - 1
        PrintN(" " + RSet(StrU(*a\ubyte[k], #PB_Byte), decimals) + " + " + RSet(StrU(*b\ubyte[k], #PB_Byte), decimals) + " = " + RSet(StrU(*gpu\ubyte[k], #PB_Byte), decimals) + ", " + RSet(StrU(*cpu\ubyte[k], #PB_Byte), decimals))
      Next k
    Case #GPUCALC_TYPE_INT32
      For k = 0 To elements - 1
        PrintN(" " + RSet(Str(*a\int[k]), decimals) + " + " + RSet(Str(*b\int[k]), decimals) + " = " + RSet(Str(*gpu\int[k]), decimals) + ", " + RSet(Str(*cpu\int[k]), decimals))
      Next k
    Case #GPUCALC_TYPE_INT16
      For k = 0 To elements - 1
        PrintN(" " + RSet(Str(*a\short[k]), decimals) + " + " + RSet(Str(*b\short[k]), decimals) + " = " + RSet(Str(*gpu\short[k]), decimals) + ", " + RSet(Str(*cpu\short[k]), decimals))
      Next k
    Case #GPUCALC_TYPE_INT8
      For k = 0 To elements - 1
        PrintN(" " + RSet(Str(*a\byte[k]), decimals) + " + " + RSet(Str(*b\byte[k]), decimals) + " = " + RSet(Str(*gpu\byte[k]), decimals) + ", " + RSet(Str(*cpu\byte[k]), decimals))
      Next k
  EndSelect
EndProcedure

Procedure.d Test(datatype.i, elements.i)
  Protected *context.GPUCalcContext = #Null
  Protected *dataset1.DATASETMEMORY = #Null
  Protected *dataset2.DATASETMEMORY = #Null
  Protected *datasetResultGPU.DATASETMEMORY = #Null
  Protected *datasetResultCPU.DATASETMEMORY = #Null
  Protected code.s
  Protected result.d = NaN()
  Protected timeNeededGPU.i
  Protected timeNeededCPU.i
  Protected difference.d
  Protected k.i
  
  *context = gpuCalcCreateContext(datatype, 8, 8)
  If *context = #Null
    ProcedureReturn NaN()
  EndIf
  
  *dataset1         = *context\allocateDatasetMemory()
  *dataset2         = *context\allocateDatasetMemory()
  *datasetResultGPU = *context\allocateDatasetMemory()
  *datasetResultCPU = *context\allocateDatasetMemory()
  
  If *dataset1 = #Null Or *dataset2 = #Null Or *datasetResultGPU = #Null Or *datasetResultCPU = #Null
    If *dataset1 <> #Null
      gpuCalcReleaseDataset(*dataset1)
    EndIf
    If *dataset2 <> #Null
      gpuCalcReleaseDataset(*dataset2)
    EndIf
    If *datasetResultGPU <> #Null
      gpuCalcReleaseDataset(*datasetResultGPU)
    EndIf
    If *datasetResultCPU <> #Null
      gpuCalcReleaseDataset(*datasetResultCPU)
    EndIf
    
    ProcedureReturn NaN()
  EndIf
  
  FillRandom(datatype, *dataset1, elements)
  FillRandom(datatype, *dataset2, elements)
  
  Select datatype
    Case #GPUCALC_TYPE_FLOAT32
      code = "vec4 calc(int startOffset) { return getDataF(0, startOffset) + getDataF(1, startOffset); }"
    Case #GPUCALC_TYPE_UINT32, #GPUCALC_TYPE_UINT16, #GPUCALC_TYPE_UINT8
      code = "uvec4 calc(int startOffset) { return getDataU(0, startOffset) + getDataU(1, startOffset); }"
    Case #GPUCALC_TYPE_INT32, #GPUCALC_TYPE_INT16, #GPUCALC_TYPE_INT8
      code = "ivec4 calc(int startOffset) { return getDataI(0, startOffset) + getDataI(1, startOffset); }"
  EndSelect
  
  If WaitForLastCommand(*context)
    *context\loadDataset(0, datatype, *dataset1, elements)
    If WaitForLastCommand(*context)
      *context\loadDataset(1, datatype, *dataset2, elements)
      If WaitForLastCommand(*context)
        
        *context\loadCode(0, code, StringByteLength(code, #PB_Ascii))
        If WaitForLastCommand(*context)
          *context\bindCode(0)
          If WaitForLastCommand(*context)
            timeNeededGPU = ElapsedMilliseconds()
            *context\calculate(*datasetResultGPU, elements)
            If WaitForLastCommand(*context)
              timeNeededGPU = ElapsedMilliseconds() - timeNeededGPU
              
              Select Datatype
                Case #GPUCALC_TYPE_FLOAT32
                  timeNeededCPU = ElapsedMilliseconds()
                  For k = 0 To elements - 1
                    *datasetResultCPU\float[k] = *dataset1\float[k] + *dataset2\float[k]
                  Next k
                  timeNeededCPU = ElapsedMilliseconds() - timeNeededCPU
                  
                Case #GPUCALC_TYPE_UINT32
                  timeNeededCPU = ElapsedMilliseconds()
                  For k = 0 To elements - 1
                    *datasetResultCPU\uint[k] = *dataset1\uint[k] + *dataset2\uint[k]
                  Next k
                  timeNeededCPU = ElapsedMilliseconds() - timeNeededCPU
                  
                Case #GPUCALC_TYPE_UINT16
                  timeNeededCPU = ElapsedMilliseconds()
                  For k = 0 To elements - 1
                    *datasetResultCPU\ushort[k] = *dataset1\ushort[k] + *dataset2\ushort[k]
                  Next k
                  timeNeededCPU = ElapsedMilliseconds() - timeNeededCPU
                  
                Case #GPUCALC_TYPE_UINT8
                  timeNeededCPU = ElapsedMilliseconds()
                  For k = 0 To elements - 1
                    *datasetResultCPU\ubyte[k] = *dataset1\ubyte[k] + *dataset2\ubyte[k]
                  Next k
                  timeNeededCPU = ElapsedMilliseconds() - timeNeededCPU
                  
                Case #GPUCALC_TYPE_INT32
                  timeNeededCPU = ElapsedMilliseconds()
                  For k = 0 To elements - 1
                    *datasetResultCPU\int[k] = *dataset1\int[k] + *dataset2\int[k]
                  Next k
                  timeNeededCPU = ElapsedMilliseconds() - timeNeededCPU
                  
                Case #GPUCALC_TYPE_INT16
                  timeNeededCPU = ElapsedMilliseconds()
                  For k = 0 To elements - 1
                    *datasetResultCPU\short[k] = Clamp(*dataset1\short[k] + *dataset2\short[k], -32768, 32767)
                  Next k
                  timeNeededCPU = ElapsedMilliseconds() - timeNeededCPU
                  
                Case #GPUCALC_TYPE_INT8
                  timeNeededCPU = ElapsedMilliseconds()
                  For k = 0 To elements - 1
                    *datasetResultCPU\byte[k] = Clamp(*dataset1\byte[k] + *dataset2\byte[k], -128, 127)
                  Next k
                  timeNeededCPU = ElapsedMilliseconds() - timeNeededCPU
              EndSelect
              
              PrintN("Time GPU: " + Str(timeNeededGPU) + "ms")
              PrintN("Time CPU: " + Str(timeNeededCPU) + "ms")
              If timeNeededCPU > 0
                PrintN("GPU/CPU:  " + StrD(timeNeededGPU / timeNeededCPU, 5))
              Else
                PrintN("GPU/CPU:  division by zero!")
              EndIf
              
              PrintN(" First 5 results: ")
              PrintFirstNResults(datatype, 5, *dataset1, *dataset2, *datasetResultGPU, *datasetResultCPU)
              
              result = CompareDatasets(datatype, elements, *datasetResultCPU, *datasetResultGPU)
            EndIf
          EndIf
        EndIf
      EndIf
    EndIf
  EndIf
  
  gpuCalcReleaseDataset(*dataset1)
  gpuCalcReleaseDataset(*dataset2)
  gpuCalcReleaseDataset(*datasetResultGPU)
  gpuCalcReleaseDataset(*datasetResultCPU)
  
  gpuCalcReleaseContext(*context)
  
  ProcedureReturn result
EndProcedure

OpenConsole()

Define *test.GPUCalcContext

; how much testdata do we need?
#DATASET_COUNT = 1024 * 1024 * 16

If LoadGPUCalcSymbols()
  *test = gpuCalcCreateContext(#GPUCALC_TYPE_FLOAT32, 8, 8)
  If *test
    If WaitForLastCommand(*test)
      If Not *test\integerCalculationSupported()
        PrintN("Only able to test float32")
        PrintN("")
        PrintN("")
        
        PrintN("Testing float32 ...")
        PrintN("Maximum difference between calculations: " + StrD(Test(#GPUCALC_TYPE_FLOAT32, #DATASET_COUNT)))
        PrintN("")
      Else
        PrintN("Testing float32 ...")
        PrintN("Maximum difference between calculations: " + StrD(Test(#GPUCALC_TYPE_FLOAT32, #DATASET_COUNT)))
        PrintN("")
        
        PrintN("Testing int32 ...")
        PrintN("Maximum difference between calculations: " + StrD(Test(#GPUCALC_TYPE_INT32, #DATASET_COUNT)))
        PrintN("")
        
        PrintN("Testing int16 ...")
        PrintN("Maximum difference between calculations: " + StrD(Test(#GPUCALC_TYPE_INT16, #DATASET_COUNT)))
        PrintN("")
        
        PrintN("Testing int8 ...")
        PrintN("Maximum difference between calculations: " + StrD(Test(#GPUCALC_TYPE_INT8, #DATASET_COUNT)))
        PrintN("")
        
        PrintN("Testing uint32 ...")
        PrintN("Maximum difference between calculations: " + StrD(Test(#GPUCALC_TYPE_UINT32, #DATASET_COUNT)))
        PrintN("")
        
        PrintN("Testing uint16 ...")
        PrintN("Maximum difference between calculations: " + StrD(Test(#GPUCALC_TYPE_UINT16, #DATASET_COUNT)))
        PrintN("")
        
        PrintN("Testing uint8 ...")
        PrintN("Maximum difference between calculations: " + StrD(Test(#GPUCALC_TYPE_UINT8, #DATASET_COUNT)))
        PrintN("")
      EndIf
    EndIf
    
    gpuCalcReleaseContext(*test)
  EndIf
  ReleaseGPUCalcSymbols()
Else
  PrintN("Error: couldn't open the library.")
EndIf
Input()

End
; IDE Options = PureBasic 4.60 Beta 3 (Windows - x64)
; ExecutableFormat = Console
; CursorPosition = 368
; FirstLine = 351
; Folding = --
; EnableThread
; EnableXP
; Executable = test32.exe